/*
 * CS560: Systems
 * 	modified from CS460 Operating Systems -- Jim Plank
 *      modified for ESIAL by Martin Quinson.
 *
 * dphil_skeleton.c -- Dining philosophers driver program
 */

#include "philosophes.h"

#include <stdio.h>
#include <pthread.h>
#include <unistd.h>
#include <math.h> /* floor */
#include <sys/time.h> /* gettimeofday */


void dsleep(double sec) {
  sleep(sec);
  (void)usleep( (sec - floor(sec)) * 1000000);
}
double dtime(void) {
  struct timeval tv;

  gettimeofday(&tv, NULL);

  return (double)(tv.tv_sec + tv.tv_usec / 1000000.0);
}

typedef struct {
  int id;                /* The philosopher's id: 0 to 5 */

  double t0;             /* The time when the program started */
  long ms;               /* The maximum time that philosopher sleeps/eats */
  double *blocktime;     /* Total time that a philosopher is blocked */
  double *blockstarting; /* If a philsopher is currently blocked, the time that he
                            started blocking */
  int phil_count;
  pthread_mutex_t *blockmon;   /* mutex for blocktime */
} Phil_struct;

int *forks; /* Who have the fork? */
pthread_mutex_t *fork_mutex; /* to protect accesses to forks array */

int errors = 0; /* amount of CS violation seen */

void *philosopher(void *v) {
  Phil_struct *ps = (Phil_struct *) v;
  double st;
  double t;

  while(1) {

    /* First the philosopher thinks for a random number of seconds */

    st = (double)((random()%ps->ms) + 1) / 10;
    printf("%3.1f Philosopher %d thinking for %.1f second%s\n",
                dtime()-ps->t0, ps->id, st, (st == 1) ? "" : "s");
    fflush(stdout);
    dsleep(st);

    /* Now, the philosopher wakes up and wants to eat.  He calls pickup
       to pick up the chopsticks */

    printf("%3.1f Philosopher %d hungry -- calling pickup()\n",
            dtime()-ps->t0, ps->id);
    fflush(stdout);
    t = dtime();
    pthread_mutex_lock(ps->blockmon);
    ps->blockstarting[ps->id] = t;
    pthread_mutex_unlock(ps->blockmon);

    /* Call user code in charge of the synchronization */
    pickup(ps->id, ps->phil_count);

    /* Check CS violations */
    pthread_mutex_lock(fork_mutex);
    if (forks[ps->id] != -1) {
    	errors++;
    	fprintf(stderr,"%3.1f The philosopher %d was granted access to fork %d even if philosopher %d already use it!! (%d errors so far)\n",
    			dtime()-ps->t0,
    			ps->id,ps->id,forks[ps->id], errors);
    }
    if (forks[(ps->id+1)%ps->phil_count] != -1) {
    	errors++;
    	fprintf(stderr,"%3.1f The philosopher %d was granted access to fork %d even if philosopher %d already use it!! (%d errors so far)\n",
    			dtime()-ps->t0,
    			ps->id,(ps->id+1)%ps->phil_count,forks[(ps->id+1)%ps->phil_count], errors);
    }
    forks[ps->id] = forks[(ps->id+1)%ps->phil_count] = ps->id;
    pthread_mutex_unlock(fork_mutex);

    pthread_mutex_lock(ps->blockmon);
    ps->blocktime[ps->id] += (dtime() - t);
    ps->blockstarting[ps->id] = -1;
    pthread_mutex_unlock(ps->blockmon);

    /* When pickup returns, the philosopher can eat for a random number of
       seconds */

    st = (double)((random()%ps->ms) + 1) /10.0;
    printf("%3.1f Philosopher %d eating for %.1f second%s\n",
                dtime()-ps->t0, ps->id, st, (st < 1) ? "" : "s");
    fflush(stdout);
    dsleep(st);

    /* Finally, the philosopher is done eating, and calls putdown to
       put down the chopsticks */

    printf("%3.1f Philosopher %d no longer eating -- calling putdown()\n",
            dtime()-ps->t0, ps->id);
    fflush(stdout);
    forks[ps->id] = forks[(ps->id+1)%ps->phil_count] = -1;
    putdown(ps->id, ps->phil_count);
  }
}

int main(int argc, char *argv[]) {
  int i;
  pthread_t *threads;
  Phil_struct *ps;
  double t0, ttmp, ttmp2;
  pthread_mutex_t *blockmon;
  double *blocktime;
  double *blockstarting;
  char s[500];
  int phil_count;
  char *curr;
  int total;
  int endtime=0;

  if (argc < 3) {
    fprintf(stderr, "usage: dphil philosopher_count maxsleepsec [endtime]\n");
    exit(1);
  }

  srandom(dtime());

  phil_count = atoi(argv[1]);
  if (argc == 4)
    endtime=atoi(argv[3]);
  threads = (pthread_t *) malloc(sizeof(pthread_t)*phil_count);
  if (threads == NULL) { perror("malloc"); exit(1); }
  ps = (Phil_struct *) malloc(sizeof(Phil_struct)*phil_count);
  if (ps == NULL) { perror("malloc"); exit(1); }

  initialize_data(phil_count);
  t0 = dtime();
  blocktime = (double *) malloc(sizeof(double)*phil_count);
  if (blocktime == NULL) { perror("malloc blocktime"); exit(1); }
  blockstarting = (double *) malloc(sizeof(double)*phil_count);
  if (blockstarting == NULL) { perror("malloc blockstarting"); exit(1); }
  forks = (int*)malloc(sizeof(int)*phil_count);
  if (forks == NULL) { perror("malloc forks"); exit(1); }

  for (i = 0; i < phil_count; i++) {
    blocktime[i] = 0;
    blockstarting[i] = -1;
    forks[i] = -1;
  }

  blockmon = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(blockmon, NULL);

  fork_mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(fork_mutex, NULL);

  for (i = 0; i < phil_count; i++) {
    ps[i].id = i;
    ps[i].t0 = t0;
    ps[i].ms = atoi(argv[2])*10;
    ps[i].blocktime = blocktime;
    ps[i].blockstarting = blockstarting;
    ps[i].blockmon = blockmon;
    ps[i].phil_count = phil_count;
    pthread_create(threads+i, NULL, philosopher, (void *) (ps+i));
  }

  while(1) {
    pthread_mutex_lock(blockmon);
    ttmp = dtime();
    curr = s;
    total = 0;
    for(i=0; i < phil_count; i++) {
      total += blocktime[i];
      if (blockstarting[i] != -1) total += (ttmp - blockstarting[i]);
    }
    sprintf(curr,"%3.1f Total blocktime: %5d : ", ttmp-t0, total);

    curr = s + strlen(s);

    for(i=0; i < phil_count; i++) {
      ttmp2 = blocktime[i];
      if (blockstarting[i] != -1) ttmp2 += (ttmp - blockstarting[i]);
      sprintf(curr, "%5.1f ", ttmp2);
      curr = s + strlen(s);
    }
    pthread_mutex_unlock(blockmon);
    printf("%s (%d CS violation)\n", s, errors);
    fflush(stdout);
    if (endtime && ttmp-t0>=endtime)
      exit(0);
    sleep(MAX(MIN(10,endtime+t0-ttmp),1));
  }
  free(ps);
  free(blocktime);
  free(forks);
}
