/*
 * pthread_intercept.c: Piece of black magic allowing to intercept
 * 		        most of the pthread and semaphore calls to
 * 		        detect deadlocks.
 * 
 *   Build the interposer as follows:
 *      gcc -shared -o pthread_interposer.so -fpic pthread_interposer.c -ldl
 *   To use it, make sure that LD_PRELOAD contains ./pthread_interposer.so
 *   For example (with bash): LD_PRELOAD=./pthread_interposer.so ./my_app
 *    or export LD_PRELOAD=./pthread_interposer.so
 *       ./my_app
 *    or setenv LD_PRELOAD "./pthread_interposer.so" (with tcsh)
 * 
 *  Author: Martin Quinson (@loria.fr)
 *  License: LGPL.
 *
 *  TODO:
 *    - implement the trylock functions
 *    - a compile-time intercepter?
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

/*
 *  Interception logic
 */
static int (*mutex_init)(pthread_mutex_t*,const pthread_mutexattr_t*) = NULL;
static int (*mutex_lock)(pthread_mutex_t*) = NULL;
static int (*mutex_trylock)(pthread_mutex_t*) = NULL;
static int (*mutex_unlock)(pthread_mutex_t*) = NULL;
static int (*mutex_destroy)(pthread_mutex_t*) = NULL;
static sem_t *(*raw_sem_open)(const char *, int) = NULL;
static int (*raw_sem_init)(sem_t*, int, unsigned int) = NULL;
static int (*raw_sem_wait)(sem_t*) = NULL;
static int (*raw_sem_post)(sem_t*) = NULL;

/*
 * Deadlock detection logic
 */

static pthread_mutex_t intermutex = PTHREAD_MUTEX_INITIALIZER;
typedef struct {
	pthread_t t; /* who am i */
	int waiting_for; /* the resource I'm locked into */
} owner_desc;
typedef struct {
	void *lock;
	int owner;
} lock_desc;

static owner_desc owners[1000];
int o_size = 0;

static lock_desc locks[10000];
int l_size=0;

static int inited = 0;

static void intercepter_init() {
        /* intercepter initialization */
	mutex_init = (int(*)(pthread_mutex_t*,const pthread_mutexattr_t*)) dlsym(RTLD_NEXT, "pthread_mutex_init");
	mutex_lock = (int(*)(pthread_mutex_t*)) dlsym(RTLD_NEXT, "pthread_mutex_lock");
	mutex_trylock = (int(*)(pthread_mutex_t*)) dlsym(RTLD_NEXT, "pthread_mutex_trylock");
	mutex_unlock = (int(*)(pthread_mutex_t*)) dlsym(RTLD_NEXT, "pthread_mutex_unlock");
	mutex_destroy = (int(*)(pthread_mutex_t*)) dlsym(RTLD_NEXT, "pthread_mutex_destroy");

	raw_sem_open = (sem_t *(*)(const char *, int)) dlsym(RTLD_NEXT, "sem_open");
	raw_sem_init = (int(*)(sem_t*, int, unsigned int)) dlsym(RTLD_NEXT, "sem_init");
	raw_sem_wait = (int(*)(sem_t*)) dlsym(RTLD_NEXT, "sem_wait");
	raw_sem_post = (int(*)(sem_t*)) dlsym(RTLD_NEXT, "sem_post");

	/* Add main thread to my data structure */
	owners[0].t = pthread_self();
	owners[0].waiting_for = -1;
	o_size = 1;
   
   inited=1;
}

/* Intercepter of pthread_create, just to make sure that we can initialize ourselves before the multithreading madness begins */
int pthread_create(pthread_t * thread, const pthread_attr_t * attr, void * (*start_routine)(void *), void * arg) {
	int res;
	static int (*func)(pthread_t *, const pthread_attr_t *, void * (*)(void *), void *);

	if (!func) {
		func = (int (*)(pthread_t *, const pthread_attr_t *, void * (*)(void *), void *)) dlsym(RTLD_NEXT, "pthread_create");
		intercepter_init();
	}
	mutex_lock(&intermutex);
	res = func(thread,attr,start_routine,arg);
	owners[o_size].t = *thread;
	owners[o_size].waiting_for = 0;
	o_size++;
	mutex_unlock(&intermutex);

	return res;
}

void* check_create(void *lock) {
	locks[l_size].lock = lock;
	locks[l_size].owner = -1;
	l_size++;
	return lock;
}
int pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *mutexattr) {
	int res;

	if (!mutex_init) intercepter_init();

	mutex_lock(&intermutex);
	res = mutex_init(mutex,mutexattr);
	check_create(mutex);
	mutex_unlock(&intermutex);

	return res;
}

static int find_lock(void*lock){
	int res;
	for (res=0;res<l_size;res++)
		if (locks[res].lock == lock)
			return res;
	fprintf(stderr," LOCK NOT FOUND \n");
	exit(1);
}
static int find_me(){
	int res;
	pthread_t me = pthread_self();

	for (res=0;res<o_size;res++)
		if (owners[res].t == me)
			return res;
	fprintf(stderr," THREAD NOT FOUND \n");
	exit(1);
}

void check_pre_lock(void *lock) {
	mutex_lock(&intermutex);
	int lock_id = find_lock(lock);
	int me = find_me();
	if (locks[lock_id].owner == -1) {
		//		fprintf(stderr,"[%d] I want l%d (free)\n",me,lock_id);
		locks[lock_id].owner = me;
	} else {
		//		int dude = locks[lock_id].owner;
		//		fprintf(stderr,"[%d] I want l%d (owned by dude %d, which waits for l%d)\n",
		//				me,lock_id, dude, owners[dude].waiting_for);
		owners[me].waiting_for = lock_id;
	}

	int dude = locks[lock_id].owner;
	do {
		if (owners[dude].waiting_for == -1)
			dude = -1; /* this guy does not wait, no deadlock (yet) */
		else
			/* one step further in the chain of waits */
			dude = locks[ owners[dude].waiting_for ].owner;
	} while (dude!= -1 && dude!=me);
	if (dude == me) {
		fprintf(stderr,"DEADLOCK DETECTED:\n");
		/* redo the loop, displaying the status */
		dude = locks[lock_id].owner;
		fprintf(stderr, "  - Thread %d waits for lock %d, owned by thread %d.\n",
				dude,owners[dude].waiting_for,locks[ owners[dude].waiting_for ].owner);
		do {
			dude = locks[ owners[dude].waiting_for ].owner;
			fprintf(stderr, "  - Thread %d waits for lock %d, owned by thread %d.\n",
					dude,owners[dude].waiting_for,locks[ owners[dude].waiting_for ].owner);
		} while (dude!=me);
	}
	mutex_unlock(&intermutex);
}

void check_post_lock(void *lock) {
	int me = find_me();
	mutex_lock(&intermutex);
	owners[me].waiting_for = -1;
	locks[find_lock(lock)].owner = me;
	mutex_unlock(&intermutex);
}

int pthread_mutex_lock(pthread_mutex_t *mutex) {
	int res;

	check_pre_lock(mutex);
	res = mutex_lock(mutex);
	check_post_lock(mutex);

	return res;
}

/*int pthread_mutex_trylock(pthread_mutex_t *mutex) {
	int res;

	res = mutex_trylock(mutex);
	fprintf(stderr,"pthread_mutex_trylock() done\n");

	return res;
}*/

void check_pre_unlock(void *lock) {
	mutex_lock(&intermutex);
	int lock_id = find_lock(lock);
	//	int me = find_me();
	//	fprintf(stderr,"[%d] unlocking %d\n",me, lock_id);
	locks[lock_id].owner = -1;
	mutex_unlock(&intermutex);
}

int pthread_mutex_unlock(pthread_mutex_t *mutex) {

	check_pre_unlock(mutex);
	return mutex_unlock(mutex);
}

int sem_init(sem_t *sem, int pshared, unsigned int value) {
	int res;

	mutex_lock(&intermutex);
	res=raw_sem_init(sem,pshared,value);
	check_create(sem);
	mutex_unlock(&intermutex);
	return res;
}

int sem_wait(sem_t *sem) {
	int res;

	check_pre_lock(sem);
	res = raw_sem_wait(sem);
	check_post_lock(sem);
	return res;
}

int sem_post(sem_t *sem) {
	check_pre_unlock(sem);
	return raw_sem_post(sem);
}
